/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/indent/src/crc.new.c                                   *
 * Created:       2000-09-26 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-08-08 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2000-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: crc.c 40 2005-12-28 21:36:26Z hampa $ */


#include "crc.h"


#define CRC12_HIGH_BIT (1U << 11)
#define CRC16_HIGH_BIT (1U << 15)
#define CRC32_HIGH_BIT (1UL << 31)


static
void crc_uint64_add (unsigned long *cnt, unsigned long n)
{
	n &= 0xffffffff;

	cnt[0] = (cnt[0] + n) & 0xffffffff;

	if (cnt[0] < n) {
		cnt[1] = (cnt[1] + 1) & 0xffffffff;
	}
}

static
void crc_uint64_shr (unsigned long *cnt, unsigned n)
{
	if (n == 0) {
		return;
	}

	if (n < 32) {
		cnt[0] = ((cnt[0] >> n) | (cnt[1] << (32 - n))) & 0xffffffff;
		cnt[1] = cnt[1] >> n;
	}
	else if (n < 64) {
		cnt[0] = cnt[1] >> (n - 32);
		cnt[1] = 0;
	}
	else {
		cnt[0] = 0;
		cnt[1] = 0;
	}
}


static
void crc12_create_tab (crc12_t *crc)
{
	unsigned       i, j;
	unsigned short reg;

	for (i = 0; i < 256; i++) {
		reg = i << 4;

		for (j = 0; j < 8; j++) {
			if (reg & CRC12_HIGH_BIT) {
				reg = (reg << 1) ^ crc->gen;
			}
			else {
				reg = reg << 1;
			}
		}

		crc->tab[i] = reg & 0x0fff;
	}
}

void crc12_init (crc12_t *crc, unsigned short gen)
{
	crc->gen = gen & 0x0fff;

	if (crc->gen == 0) {
		crc->gen = CRC_CRC12;
	}

	crc->reg = 0;
	crc->cnt[0] = 0;
	crc->cnt[1] = 0;
	crc12_create_tab (crc);
}

void crc12_reset (crc12_t *crc)
{
	crc->reg = 0;
	crc->cnt[0] = 0;
	crc->cnt[1] = 0;
}

void crc12_copy (crc12_t *dst, const crc12_t *src)
{
	*dst = *src;
}

unsigned short crc12_get_reg (const crc12_t *crc)
{
	return (crc->reg);
}

void crc12_calc (crc12_t *crc, const void *buf, unsigned long cnt)
{
	unsigned            i;
	const unsigned char *src = buf;

	crc_uint64_add (crc->cnt, cnt);

	while (cnt > 0) {
		i = (crc->reg >> 4) & 0xff;

		crc->reg = (((crc->reg << 8) | *src) ^ crc->tab[i]) & 0x0fff;

		src += 1;
		cnt -= 1;
	}
}

void crc12_done (crc12_t *crc, int posix)
{
	unsigned      i;
	unsigned long cnt[2];
	unsigned char buf;

	if (posix) {
		cnt[0] = crc->cnt[0];
		cnt[1] = crc->cnt[1];

		while ((cnt[0] != 0) || (cnt[1] != 0)) {
			buf = cnt[0] & 0xff;
			crc12_calc (crc, &buf, 1);
			crc_uint64_shr (cnt, 8);
		}
	}

	for (i = 0; i < 12; i++) {
		if (crc->reg & CRC12_HIGH_BIT) {
			crc->reg = ((crc->reg << 1) ^ crc->gen) & 0x0fff;
		}
		else {
			crc->reg = (crc->reg << 1) & 0x0fff;
		}
	}

	if (posix) {
		crc->reg = crc->reg ^ 0x0fff;
	}
}


static
void crc16_create_tab (crc16_t *crc)
{
	unsigned       i, j;
	unsigned short reg;

	for (i = 0; i < 256; i++) {
		reg = i << 8;

		for (j = 0; j < 8; j++) {
			if (reg & CRC16_HIGH_BIT) {
				reg = (reg << 1) ^ crc->gen;
			}
			else {
				reg = reg << 1;
			}
		}

		crc->tab[i] = reg & 0xffff;
	}
}

void crc16_init (crc16_t *crc, unsigned short gen)
{
	crc->gen = gen & 0xffff;

	if (crc->gen == 0) {
		crc->gen = CRC_CRC16;
	}

	crc->reg = 0;
	crc->cnt[0] = 0;
	crc->cnt[1] = 0;
	crc16_create_tab (crc);
}

void crc16_reset (crc16_t *crc)
{
	crc->reg = 0;
	crc->cnt[0] = 0;
	crc->cnt[1] = 0;
}

void crc16_copy (crc16_t *dst, const crc16_t *src)
{
	*dst = *src;
}

unsigned short crc16_get_reg (const crc16_t *crc)
{
	return (crc->reg);
}

void crc16_calc (crc16_t *crc, const void *buf, unsigned long cnt)
{
	unsigned            i;
	const unsigned char *src = buf;

	crc_uint64_add (crc->cnt, cnt);

	while (cnt > 0) {
		i = (crc->reg >> 8) & 0xff;

		crc->reg = (((crc->reg << 8) | *src) ^ crc->tab[i]) & 0xffff;

		src += 1;
		cnt -= 1;
	}
}

void crc16_done (crc16_t *crc, int posix)
{
	unsigned long cnt[2];
	unsigned char buf[2];

	if (posix) {
		cnt[0] = crc->cnt[0];
		cnt[1] = crc->cnt[1];

		while ((cnt[0] != 0) || (cnt[1] != 0)) {
			buf[0] = cnt[0] & 0xff;
			crc16_calc (crc, buf, 1);
			crc_uint64_shr (cnt, 8);
		}
	}

	buf[0] = 0;
	buf[1] = 0;
	crc16_calc (crc, buf, 2);

	if (posix) {
		crc->reg = crc->reg ^ 0xffff;
	}
}


static
void crc32_create_tab (crc32_t *crc)
{
	unsigned     i, j;
	unsigned long reg;

	for (i = 0; i < 256; i++) {
		reg = (unsigned long) i << 24;

		for (j = 0; j < 8; j++) {
			if (reg & CRC32_HIGH_BIT) {
				reg = (reg << 1) ^ crc->gen;
			}
			else {
				reg = reg << 1;
			}
		}

		crc->tab[i] = reg & 0xffffffff;
	}
}

void crc32_init (crc32_t *crc, unsigned long gen)
{
	crc->gen = gen & 0xffffffff;

	if (crc->gen == 0) {
		crc->gen = CRC_CRC32;
	}

	crc->reg = 0;
	crc->cnt[0] = 0;
	crc->cnt[1] = 0;
	crc32_create_tab (crc);
}

void crc32_reset (crc32_t *crc)
{
	crc->reg = 0;
	crc->cnt[0] = 0;
	crc->cnt[1] = 0;
}

void crc32_copy (crc32_t *dst, const crc32_t *src)
{
	*dst = *src;
}

unsigned long crc32_get_reg (const crc32_t *crc)
{
	return (crc->reg);
}

void crc32_calc (crc32_t *crc, const void *buf, unsigned long cnt)
{
	unsigned            i;
	const unsigned char *src = buf;

	crc_uint64_add (crc->cnt, cnt);

	while (cnt > 0) {
		i = (crc->reg >> 24) & 0xff;

		crc->reg = (((crc->reg << 8) | *src) ^ crc->tab[i]) & 0xffffffff;

		src += 1;
		cnt -= 1;
	}
}

void crc32_done (crc32_t *crc, int posix)
{
	unsigned long cnt[2];
	unsigned char buf[4];

	if (posix) {
		cnt[0] = crc->cnt[0];
		cnt[1] = crc->cnt[1];

		while ((cnt[0] != 0) || (cnt[1] != 0)) {
			buf[0] = cnt[0] & 0xff;
			crc32_calc (crc, buf, 1);
			crc_uint64_shr (cnt, 8);
		}
	}

	buf[0] = 0;
	buf[1] = 0;
	buf[2] = 0;
	buf[3] = 0;
	crc32_calc (crc, buf, 4);

	if (posix) {
		crc->reg = crc->reg ^ 0xffffffff;
	}
}
