/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lecdec.c                                               *
 * Created:       2005-08-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: lecdec.c 64 2006-11-30 16:15:58Z hampa $ */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lstdint.h"

#include "leco.h"
#include "block.h"
#include "crc.h"
#include "gf.h"
#include "lecdec.h"
#include "lecofile.h"
#include "matrix.h"
#include "md5.h"
#include "util.h"


typedef struct {
	lec_file_t    lf;

	FILE          *inp;
	FILE          *out;

	unsigned      bcntok;
	unsigned      ecntok;

	unsigned char *bmsk;
	unsigned long *bmap;

	uint32_t      *crc;

	unsigned char *buf1;
	unsigned char *buf2;

	uint16_t      *blk1;
	uint16_t      *blk2;

	uint16_t      *G;
	uint16_t      *GI;
} lec_dec_t;


int lec_dec_open (lec_dec_t *ld, FILE *inp, FILE *lec, FILE *out)
{
	lec_file_t *lf;

	lf = &ld->lf;

	if (lecf_open (&ld->lf, lec)) {
		printf ("error opening leco file\n");
		return (1);
	}

	if (ld->lf.base != 0) {
		printf ("base:             0x%08llx\n",
			(unsigned long long) ld->lf.base
		);
	}

	if (ld->lf.hofs != 0) {
		printf ("header offset:    0x%08llx\n",
			(unsigned long long) ld->lf.hofs
		);
	}

	printf ("block size:       %lu (%lu * %u bits)\n",
		lf->size, lf->bsize, lf->wsize
	);
	printf ("data blocks:      %u\n", lf->bcnt);
	printf ("ecc blocks:       %u\n", lf->ecnt);
	printf ("file size:        %llu\n", (unsigned long long) lf->fsize);

	ld->inp = inp;
	ld->out = out;

	ld->bcntok = 0;
	ld->ecntok = 0;

	ld->bmsk = NULL;
	ld->bmap = NULL;

	ld->crc = NULL;

	ld->buf1 = NULL;
	ld->buf2 = NULL;
	ld->blk1 = NULL;
	ld->blk2 = NULL;

	ld->G = NULL;
	ld->GI = NULL;

	return (0);
}

void lec_dec_free (lec_dec_t *ld)
{
	free (ld->bmsk);
	free (ld->crc);
	free (ld->buf1);
	free (ld->blk1);
	free (ld->G);
	free (ld->GI);
}

int lec_dec_alloc (lec_dec_t *ld)
{
	ld->buf1 = malloc (2UL * (ld->lf.size + 8));
	if (ld->buf1 == NULL) {
		return (1);
	}

	ld->buf2 = ld->buf1 + ld->lf.size + 8;

	ld->blk1 = malloc (2UL * ld->lf.bsize * sizeof (uint16_t));
	if (ld->blk1 == NULL) {
		return (1);
	}

	ld->blk2 = ld->blk1 + ld->lf.bsize;

	return (0);
}

int lec_dec_get_bmsk (lec_dec_t *ld)
{
	unsigned      i, j, n;
	unsigned      x;
	unsigned char buf[4];
	uint16_t      *t;
	uint32_t      crc;
	lec_file_t    *lf;

	lf = &ld->lf;

	ld->G = malloc ((unsigned long) lf->bcnt * lf->bcnt * sizeof (uint16_t));
	if (ld->G == NULL) {
		return (1);
	}

	ld->crc = malloc (lf->bcnt * sizeof (uint32_t));
	if (ld->crc == NULL) {
		return (1);
	}

	ld->bmsk = malloc (lf->bcnt);
	if (ld->bmsk == NULL) {
		return (1);
	}

	ld->bmap = malloc (lf->bcnt * sizeof (unsigned long));
	if (ld->bmap == NULL) {
		return (1);
	}

	ld->bcntok = 0;

	for (i = 0; i < lf->bcnt; i++) {
		if (lecf_read (lf, lf->dofs + 4UL * i, buf, 4) != 4) {
			return (1);
		}

		ld->crc[i] = lec_get_uint32 (buf, 0);

		n = lec_read (ld->inp, (uint64_t) i * lf->size, ld->buf1, lf->size);

		for (j = n; j < lf->size; j++) {
			ld->buf1[j] = 0;
		}

		crc = lec_get_crc (ld->buf1, lf->size);

		if (ld->crc[i] == crc) {
			ld->bmsk[i] = 0;
			ld->bmap[i] = i;

			t = ld->G + i * lf->bcnt;

			for (x = 0; x < lf->bcnt; x++) {
				*(t++) = (x == i) ? 1 : 0;
			}

			ld->bcntok += 1;
		}
		else {
			ld->bmsk[i] = 1;
			ld->bmap[i] = 0;

			printf ("  bad block: data %u [%08llx - %08llx]\n",
				i,
				(unsigned long long) i * lf->size,
				(unsigned long long) i * lf->size + 2UL * lf->size - 1
			);
		}
	}

	printf ("bad blocks data:  %u of %u\n", (lf->bcnt - ld->bcntok), lf->bcnt);

	return (0);
}

int lec_dec_get_emsk (lec_dec_t *ld)
{
	unsigned   i;
	unsigned   x, y;
	unsigned   blki;
	uint16_t   *t;
	uint64_t   ofs;
	uint32_t   crc1, crc2;
	lec_file_t *lf;

	lf = &ld->lf;

	y = 0;

	ld->ecntok = 0;

	ofs = lf->eofs;

	for (i = 0; i < lf->ecnt; i++) {
		if (lecf_read (lf, ofs, ld->buf1, lf->size + 8) != (lf->size + 8)) {
			return (1);
		}

		crc2 = lec_get_crc (ld->buf1, lf->size + 4);

		blki = lec_get_uint32 (ld->buf1, 0);
		crc1 = lec_get_uint32 (ld->buf1, lf->size + 4);

		if (crc1 == crc2) {
			ld->ecntok += 1;

			/* find next missing data block */
			while ((y < lf->bcnt) && (ld->bmsk[y] == 0)) {
				y += 1;
			}

			if (y < lf->bcnt) {
				ld->bmap[y] = lf->bcnt + i;

				t = ld->G + y * lf->bcnt;

				for (x = 0; x < lf->bcnt; x++) {
					*(t++) = lec_get_coeff (blki, x);
				}

				y += 1;
			}
		}
		else {
			printf ("  bad block: ecc %u\n", i);
		}

		ofs += lf->size + 8;
	}

	printf ("bad blocks ecc:   %u of %u\n", (lf->ecnt - ld->ecntok), lf->ecnt);

	return (0);
}

/* load block i int buf and blk. use bmap. */
static
int lec_dec_load_block (lec_dec_t *ld, uint16_t *blk, void *buf, unsigned i)
{
	lec_file_t    *lf;
	uint64_t      ofs;
	unsigned long cnt;

	lf = &ld->lf;

	if (ld->bmap[i] < lf->bcnt) {
		ofs = (uint64_t) ld->bmap[i] * lf->size;
		cnt = lec_read (ld->inp, ofs, buf, lf->size);
	}
	else {
		ofs = lf->eofs + (uint64_t) (ld->bmap[i] - lf->bcnt) * (lf->size + 8);
		cnt = lecf_read (lf, ofs + 4, buf, lf->size);
	}

	if (cnt == 0) {
		return (1);
	}

	if (cnt < lf->size) {
		memset ((char *) buf + cnt, 0, lf->size - cnt);
	}

	lec_block_get (blk, buf, lf->bsize);

	return (0);
}

/* invert ld->G into ld->GI */
int lec_dec_invert (lec_dec_t *ld)
{
	lec_file_t *lf;

	lf = &ld->lf;

	ld->GI = malloc ((unsigned long) lf->bcnt * lf->bcnt * sizeof (uint16_t));
	if (ld->GI == NULL) {
		return (1);
	}

	mat_invert (ld->GI, ld->G, lf->bcnt);

	free (ld->G);
	ld->G = NULL;

	return (0);
}

/* restore blocks using ld->GI */
int lec_dec_restore (lec_dec_t *ld)
{
	unsigned      i, j;
	unsigned long cnt;
	uint32_t      crc;
	uint64_t      wofs;
	md5_t         md5;
	lec_file_t    *lf;

	lf = &ld->lf;

	md5_init (&md5);

	wofs = 0;

	for (i = 0; i < lf->bcnt; i++) {
		if (ld->bmsk[i]) {
			printf ("restoring block %u of %u  \r", i, lf->bcnt);
			fflush (stdout);

			lec_block_clear (ld->blk1, lf->bsize);

			for (j = 0; j < lf->bcnt; j++) {
				if (lec_dec_load_block (ld, ld->blk2, ld->buf2, j)) {
					return (1);
				}

				lec_block_muladd (ld->blk1, ld->blk2, ld->GI[i * lf->bcnt + j], lf->bsize);
			}

			lec_block_set (ld->buf1, ld->blk1, lf->bsize);

			crc = lec_get_crc (ld->buf1, lf->size);

			printf ("  restored block: data %u [%08llx - %08llx] (crc %s)\n",
				i,
				(unsigned long long) wofs,
				(unsigned long long) (wofs + lf->size - 1),
				(ld->crc[i] == crc) ? "ok" : "failed"
			);
			fflush (stdout);

			cnt = lf->size;
		}
		else {
			cnt = lec_read (ld->inp, wofs, ld->buf1, lf->size);
		}

		if ((i + 1) == lf->bcnt) {
			/* last block */
			unsigned long tmp;

			tmp = lf->fsize % lf->size;

			if (tmp == 0) {
				tmp = lf->size;
			}

			if (tmp < cnt) {
				cnt = tmp;
			}
		}

		md5_calc (&md5, ld->buf1, cnt);

		if (lec_write (ld->out, wofs, ld->buf1, cnt) != cnt) {
			return (1);
		}

		wofs += lf->size;
	}

	md5_done (&md5);

	md5_get_sum (&md5, ld->buf1);

	if (memcmp (ld->buf1, lf->md5, 16) == 0) {
		printf ("checking md5:     ok\n");
	}
	else {
		printf ("checking md5:     failed\n");
		return (1);
	}

	return (0);
}

int lec_dec_ld (lec_dec_t *ld)
{
	unsigned long gcnt;

	if (lec_dec_alloc (ld)) {
		return (1);
	}

	if (lec_dec_get_bmsk (ld)) {
		return (1);
	}

	if (lec_dec_get_emsk (ld)) {
		return (1);
	}

	gcnt = (unsigned long) ld->bcntok + ld->ecntok;

	if (gcnt < ld->lf.bcnt) {
		printf ("missing %lu blocks\n", ld->lf.bcnt - gcnt);
		return (1);
	}
	else {
		printf ("spare blocks:     %lu\n", gcnt - ld->lf.bcnt);
	}

	if (lec_dec_invert (ld)) {
		return (1);
	}

	if (lec_dec_restore (ld)) {
		return (1);
	}

	fflush (ld->out);

	return (0);
}

int lec_dec_fp (FILE *inp, FILE *lec, FILE *out)
{
	int       r;
	lec_dec_t ld;

	if (lec_dec_open (&ld, inp, lec, out)) {
		return (1);
	}

	r = lec_dec_ld (&ld);

	fflush (stdout);

	lec_dec_free (&ld);

	return (r);
}

int lec_decode (const char *finp, const char *flec, const char *fout)
{
	int  r;
	FILE *inp, *lec, *out;

	inp = fopen (finp, "rb");
	if (inp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
		return (1);
	}

	lec = fopen (flec, "rb");
	if (lec == NULL) {
		fprintf (stderr, "%s: can't open ecc input file (%s)\n", argv0, flec);
		fclose (inp);
		return (1);
	}

	out = fopen (fout, "w+b");
	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
		fclose (lec);
		fclose (inp);
		return (1);
	}

	r = lec_dec_fp (inp, lec, out);

	fclose (out);
	fclose (lec);
	fclose (inp);

	return (r);
}
