/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/lecofile.h                                             *
 * Created:       2005-08-11 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: lecofile.h 64 2006-11-30 16:15:58Z hampa $ */


#ifndef LECO_LECOFILE_H
#define LECO_LECOFILE_H 1


#include "config.h"

#include <stdio.h>
#include "lstdint.h"


#define LECO_MAGIC  0x4c45434f
#define LECO_MAGIC1 "[DATA]  "
#define LECO_MAGIC2 "[ECC ]  "
#define LECO_MAGIC3 "[RDNT]  "
#define LECO_MAGIC4 "[LEND]  "
#define LECO_HEADER 80


typedef struct {
	FILE               *fp;

	unsigned long long base;

	unsigned char      header[LECO_HEADER];

	unsigned long      magic;
	unsigned           version;
	unsigned           hsize;
	unsigned long      bsize;
	unsigned long      size;
	unsigned           wsize;
	unsigned long long hofs;
	unsigned long long dofs;
	unsigned long long eofs;
	unsigned long long rofs;
	unsigned long long zofs;
	unsigned           bcnt;
	unsigned           ecnt;
	unsigned long long fsize;
	unsigned char      md5[16];

	unsigned long      fcrc;
	unsigned long      dcrc;
	unsigned long      hcrc;
} lec_file_t;


/*****************************************************************************
 * @short  Get the file position
 * @param  fp  The file
 * @retval pos The file position
 * @return Zero if successful, non-zero otherwise
 *****************************************************************************/
int lec_get_pos (FILE *fp, uint64_t *pos);

/*****************************************************************************
 * @short  Set the file position
 * @param  fp  The file
 * @param  pos The file position
 * @return Zero if successful, non-zero otherwise
 *****************************************************************************/
int lec_set_pos (FILE *fp, uint64_t pos);

/*****************************************************************************
 * @short  Get the file size
 * @param  fp   The file
 * @retval size The file size
 * @return Zero if successful, non-zero otherwise
 *****************************************************************************/
int lec_get_fsize (FILE *fp, uint64_t *size);

/*****************************************************************************
 * @short  Read from a file
 * @param  fp The file
 * @param  pos The file position
 * @param  buf The read buffer
 * @param  cnt The number of bytes to read
 * @return The number of bytes read
 *****************************************************************************/
unsigned long lec_read (FILE *fp, uint64_t pos, void *buf, unsigned long cnt);

/*****************************************************************************
 * @short  Write to a file
 * @param  lf The file
 * @param  pos The file position
 * @param  buf The write buffer
 * @param  cnt The number of bytes to write
 * @return The number of bytes written
 *****************************************************************************/
unsigned long lec_write (FILE *fp, uint64_t pos, const void *buf, unsigned long cnt);

/*****************************************************************************
 * @short  Read from a file and allocate buffer
 * @param  fp The file
 * @param  pos The file position
 * @param  cnt The number of bytes to read
 * @return A pointer to a newly allocated buffer of cnt bytes or NULL on
 *         error
 *****************************************************************************/
unsigned char *lec_read_buf (FILE *fp, uint64_t pos, unsigned long cnt);

uint16_t lec_get_uint16 (const void *buf, unsigned i);
uint32_t lec_get_uint32 (const void *buf, unsigned i);
uint64_t lec_get_uint64 (const void *buf, unsigned i);

void lec_set_uint16 (void *buf, unsigned i, uint16_t val);
void lec_set_uint32 (void *buf, unsigned i, uint32_t val);
void lec_set_uint64 (void *buf, unsigned i, uint64_t val);

uint32_t lec_get_crc (const void *buf, unsigned long cnt);


/*****************************************************************************
 * @short  Initialize a leco file
 * @param  lf The leco file
 * @param  fp The file
 *****************************************************************************/
void lecf_init (lec_file_t *lf, FILE *fp);

/*****************************************************************************
 * @short  Open a leco file
 * @param  lf The leco file
 * @param  fp The file
 * @return Zero if successful, non-zero otherwise
 *****************************************************************************/
int lecf_open (lec_file_t *lf, FILE *fp);

/*****************************************************************************
 * @short  Read from a leco file
 * @param  lf The leco file
 * @param  pos The file position
 * @param  buf The read buffer
 * @param  cnt The number of bytes to read
 * @return The number of bytes read
 *****************************************************************************/
unsigned long lecf_read (lec_file_t *lf, uint64_t pos, void *buf, unsigned long cnt);

/*****************************************************************************
 * @short  Write to a leco file
 * @param  lf The leco file
 * @param  pos The file position
 * @param  buf The write buffer
 * @param  cnt The number of bytes to write
 * @return The number of bytes written
 *****************************************************************************/
unsigned long lecf_write (lec_file_t *lf, uint64_t pos, const void *buf, unsigned long cnt);

unsigned char *lecf_read_buf (lec_file_t *lf, uint64_t pos, unsigned long cnt);

int lecf_scan (lec_file_t *lf, const void *src, unsigned cnt, uint64_t *ofs);

void lecf_calc_dofs (lec_file_t *lf);
void lecf_calc_eofs (lec_file_t *lf);
void lecf_calc_rofs (lec_file_t *lf);

int lecf_write_header (lec_file_t *lf);
int lecf_write_magic1 (lec_file_t *lf);
int lecf_write_magic2 (lec_file_t *lf);
int lecf_write_magic3 (lec_file_t *lf);
int lecf_write_magic4 (lec_file_t *lf);
int lecf_write_redundant (lec_file_t *lf, const void *crc);


#endif
