/*****************************************************************************
 * leco                                                                      *
 *****************************************************************************/

/*****************************************************************************
 * File name:     src/matrix.c                                               *
 * Created:       2005-08-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2006-07-08 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2005-2006 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: matrix.c 48 2006-07-08 11:37:23Z hampa $ */


#include "config.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "lstdint.h"

#include "block.h"
#include "gf.h"
#include "lecofile.h"
#include "matrix.h"
#include "util.h"


#define MAT_USE_LR 1


/*
 * row[di] = row[di] - v * row[si], starting at column si + 1
 * b[di] = b[di] - v * b[si]
 */
static
void mat_mulsub (uint16_t *mat, uint16_t *b, unsigned n, unsigned di, unsigned si, uint16_t v)
{
	unsigned i;
	uint16_t *p1, *p2;

	p1 = mat + di * n;
	p2 = mat + si * n;

	for (i = si + 1; i < n; i++) {
		p1[i] ^= gf_mul (&par_gf, p2[i], v);
	}

	if (b != NULL) {
		b[di] ^= gf_mul (&par_gf, b[si], v);
	}
}

#if MAT_USE_LR
/*
 * solve mat * x = b using L/R decomposition
 * if (first == 0) then mat is alread decomposed
 */
void mat_solve_lr (uint16_t *mat, uint16_t *x, uint16_t *b, unsigned n, int first)
{
	unsigned i, j;
	uint16_t *p1, *p2;
	uint16_t f;
	uint16_t *c;

	if (first) {
		/* decompose mat intoe L and R */
		for (i = 0; i < n; i++) {
			p1 = mat + i * n + i;
			p2 = p1;

			assert (*p1 != 0);

			for (j = i + 1; j < n; j++) {
				p2 += n;

				if (*p2 != 0) {
					f = gf_div (&par_gf, *p2, *p1);
					mat_mulsub (mat, NULL, n, j, i, f);
					*p2 = f;
				}
			}
		}
	}

	/* L and R are in mat */

	c = malloc (n * sizeof (uint16_t));

	/* solve L * c = b */
	for (i = 0; i < n; i++) {
		c[i] = b[i];
		for (j = 0; j < i; j++) {
			c[i] ^= gf_mul (&par_gf, c[j], mat[n * i + j]);
		}
	}

	/* solve R * x = c */
	i = n;
	while (i > 0) {
		i -= 1;

		x[i] = c[i];
		for (j = i + 1; j < n; j++) {
			x[i] ^= gf_mul (&par_gf, x[j], mat[n * i + j]);
		}
		x[i] = gf_div (&par_gf, x[i], mat[n * i + i]);
	}

	free (c);
}

#else

/* solve mat * x = b using the gauss algorithm */
void mat_solve (uint16_t *mat, uint16_t *x, uint16_t *b, unsigned n)
{
	unsigned i, j;
	uint16_t v;
	uint16_t *p1, *p2;

	for (i = 0; i < n; i++) {
		p1 = mat + i * n + i;
		p2 = p1;

		assert (*p1 != 0);

		for (j = i + 1; j < n; j++) {
			p2 += n;

			if (*p2 != 0) {
				mat_mulsub (mat, b, n, j, i, gf_div (&par_gf, *p2, *p1));
			}
		}
	}

	i = n;
	while (i > 0) {
		i -= 1;

		v = b[i];

		p1 = mat + i * n + i;
		p2 = p1;

		for (j = i + 1; j < n; j++) {
			p2 += 1;
			v ^= gf_mul (&par_gf, *p2, x[j]);
		}

		x[i] = gf_div (&par_gf, v, *p1);
	}
}
#endif

/* dst = src^-1 */
void mat_invert (uint16_t *dst, const uint16_t *src, unsigned n)
{
	unsigned      i, j;
	unsigned      pct1, pct2;
	unsigned long m;
	uint16_t      *x, *b, *A;

	m = (unsigned long) n * (unsigned long) n;

	A = malloc ((m + 2UL * n) * sizeof (uint16_t));
	x = A + m;
	b = x + n;

#if MAT_USE_LR
	memcpy (A, src, m * sizeof (uint16_t));
#endif

	pct1 = 1001;

	for (i = 0; i < n; i++) {
		pct2 = (1000UL * i + n / 2) / n;
		if (pct1 != pct2) {
			printf ("inverting matrix: %u.%u%%\r", pct2 / 10, pct2 % 10);
			fflush (stdout);
			pct1 = pct2;
		}

		for (j = 0; j < n; j++) {
			b[j] = (j == i) ? 1 : 0;
		}

#if MAT_USE_LR
		mat_solve_lr (A, x, b, n, i == 0);
#else
		memcpy (A, src, m * sizeof (uint16_t));
		mat_solve (A, x, b, n);
#endif

		for (j = 0; j < n; j++) {
			dst[j * n + i] = x[j];
		}
	}

	printf ("inverting matrix: 100.0%%  \n");
	fflush (stdout);

	free (A);
}

/* dst = s1 * s2 */
void mat_mul (uint16_t *dst, uint16_t *s1, uint16_t *s2, unsigned n)
{
	unsigned      i, j, k;
	unsigned long m;
	uint16_t      *tmp;

	m = (unsigned long) n * (unsigned long) n;

	if ((dst == s1) || (dst == s2)) {
		tmp = malloc (m * sizeof (uint16_t));
	}
	else {
		tmp = dst;
	}

	for (j = 0; j < n; j++) {
		for (i = 0; i < n; i++) {
			tmp[j * n + i] = 0;

			for (k = 0; k < n; k++) {
				tmp[j * n + i] ^= gf_mul (&par_gf, s1[j * n + k], s2[k * n + i]);
			}
		}
	}

	if (tmp != dst) {
		memcpy (dst, tmp, m * sizeof (uint16_t));
		free (tmp);
	}
}
